classdef bSplines < handle
    properties 
      ordre
      coefficient
      coefficient_robust
      covariance_coefficient_robust
      knots
      observation
      valeur_gcv
      mat_inter
      valeur_k;
    end    
    methods


        function bS =  bSplines(obs,p,val_ini,total_motion)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%DESCRIPTION: % Create a B Splines object: a B-spline  estimator of the curve (i.e a piecewise polynomial estimator) directly from the observation
%
%USAGE:  bS =  bSplines(obs,p,val_ini,total_motion)
%INPUT :
%obs                        :(Observation object) observation object containing the observation and the corresponding observation time
%OPTIONAL:
%p                         :(integer default value 3)   polynomial order of the spline basis components
%val_ini                   :(d sized vector) value of the IC (if known)
%total_motion              :(d sized vector) value of the integral (if known)
%OUTPUT:
%bS                         :(bSplines Object): return the constructed the bSpline estimator of the curve               
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
            bS.observation = obs;
            T = obs.getTimes();
            
            if nargin <= 1
                p = 2+1;
            end
            bS.ordre = p;
            
          if nargin <=2
                
            [C_ D_ E_] = coeff_splines_loc(T,getMatrix(obs),p);
            
            bS.coefficient = E_;
            bS.knots = C_;
            bS.valeur_gcv = D_;
            
            bS.mat_inter = cell(1,length(C_));
            for i =1 : length(C_)
            Bn  = bsplineM(T{i},C_{1,i},p +1,0,0);
            bS.mat_inter(i) = {Bn};
            end
            
          else
              
            if nargin < 4 || isempty(total_motion)
                total_motion = cell(1,length(getMatrix(obs)));
            end

            [C_ D_ E_] = coeff_splines_loc_constraint(T,getMatrix(obs),p,val_ini,total_motion);
            
            bS.coefficient = E_;
            bS.knots = C_;
            bS.valeur_gcv = D_;
            
            bS.mat_inter = cell(1,length(C_));
            for i =1 : length(C_)
            Bn  = bsplineM(T{i},C_{1,i},p +1,0,0);
            bS.mat_inter(i) = {Bn};
            end
                
          end
            bS.valeur_k = -ones(length(C_),1);
        end
        
         % Retourne l'objet Observation li au lisseur 
        % Argument : BSplines, base de regression
        function obs = getObservation(bs)

            obs =  bs.observation;
        end
        
        % Retourne l'ensemble des noeuds dansla base de regression pour chaque dimension des observations sous forme d'un cell arrays
        % Argument : BSplines, base de regression
        function bd = getKnots(bs)
            bd = bs.knots;            
        end
        
        % Retourne l'ensemble des coefficients dansla base de regression pour chaque dimension des observations sous forme d'un cell arrays
        % Argument : BSplines, base de regression
        function coeff = getCoefficients(bs)
            coeff = bs.coefficient;
        end
        
         % Retourne l'ensemble des valeurs de gcv pour les emplacements et
         % le nombre de noeud actuel pour chaque variable observ  ( sous
         % forme d'un cell arrays)
        % Argument : BSplines, base de regression
        function val_gcv = getValeur_gcv(bs)
            val_gcv = bs.valeur_gcv;
        end
        
         % Retourne l'ensemble des valeurs de Bn (sous forme d'un cell array)
        % Argument : Bs, base de regression
        function val_bn = getBn(bs)
            val_bn = bs.mat_inter;
        end
        

        function res = calcul_prediction(r,X,order,variable)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%DESCRIPTION: % Return a non-parametric estimation of derivative form of order "order" evaluated at timesteps contained in time_vect
%
%USAGE:   res = calcul_prediction(bs,time_vect,order_der,type_fun,variable)
%
%INPUT :
%bs               :(bSplines  object) , embed the nonparametric estimators
%time_vect         :(vector) embed the different evaluation timestep
%OPTIONAL
%order_der           :(integer) derivative order (default 0)
%variable        :(integer vector) state variables indexes for which the prediction is computed (default all variables)
%OUTPUT:
%res             :(matrix of size  length(time_vect)*length(variable)) embeding in each cell the values for the corresponding state variable estimation
%               
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
        if nargin <= 2
          order = 0;
        end
        
        M = r.observation.getMatrix();
        longueur = length(M);
        
        if nargin < 4
            res = zeros(length(X),longueur);
            
        for i = 1 : longueur
        res(:,i) = bsplineM(X,r.knots{i},r.ordre+1,order,0)*r.coefficient{i};
        end
        
        else

            res = zeros(length(X),length(variable));
            for j = 1 : length(variable)
            res(:,j) = bsplineM(X,r.knots{variable(j)},r.ordre+1,order,0)*r.coefficient{variable(j)};
            end
        end

        end
        
        function res = calcul_prediction_robust(r,X,order_der,type_fun,variable)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%DESCRIPTION: % Return a robust non-parametric estimation of derivative form of order "order" evaluated at timesteps contained in time_vect
%
%USAGE:   res = calcul_prediction_robust(bs,time_vect,order_der,type_fun,variable)
%
%INPUT :
%bs               :(bSplines  object) , embed the nonparametric estimators
%time_vect         :(vector) embed the different evaluation timestep
%OPTIONAL
%order_der           :(integer) derivative order (default 0)
%type_fun            : (integer) indicate the used weighting function 
%                         1 (default) huber: w = 1 ./ max(1, abs(r))
%                         2           bisquare: w = (abs(r)<1) .* (1 - r.^2).^2
%                         3           fair:  w = 1 ./ (1 + abs(r))
%                         4           talwar:  w = 1 * (abs(r)<1)
%variable        :(integer vector) state variables indexes for which the prediction is computed (default all variables)
%OUTPUT:
%res             :(matrix length(X)*length(variable)) embeding in each cell the values for the corresponding state variable estimation
%               
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
      
            if nargin <= 2
                order_der = 0;
            end
            
            
            Cell_knots = r.knots;
            Observation = r.getObservation();
            Cell_obs = Observation.getMatrix();
            Times_obs = Observation.getTimes();
            
             if isempty(r.coefficient_robust)                    
                  cell_coeff_robust = cell(length(Times_obs));                   
             end
            
            
            if nargin <= 4
                variable = 1 : length(Cell_knots);
            end
            
            if nargin <=3
                type_fun =1;
            end
  
             res = zeros(length(X),length(variable));
                        
            for var_i = 1 : length(variable)
                knots_i = Cell_knots{variable(var_i)};
                Yi = Cell_obs{variable(var_i)};
                Ti = Times_obs{variable(var_i)};
                
                b_base = bsplineM(Ti,knots_i,r.ordre+1,0,0);
                
                     
                if type_fun == 1;
                 func_weight = 'huber';
                  tune_const = 1.345;
                end
                if type_fun ==2
                 func_weight = 'bisquare';
                  tune_const = 4.685;
                end
                if type_fun == 3
                 func_weight = 'fair';
                 tune_const = 1.400;
                end
                if type_fun ==4
                 func_weight = 'talwar';
                 tune_const = 2.795;
                end


                if isempty(r.coefficient_robust)
                    
                    [coeff_alpha,stats]= robustfit(b_base,Yi,func_weight,tune_const,'off');
                    cell_coeff_robust{var_i} = coeff_alpha;
                    covariance_matrix{var_i} = stats.covb;
                else
                    coeff_alpha = r.coefficient_robust{var_i};
                end
                res(:,var_i) = bsplineM(X,knots_i,r.ordre+1,order_der,0)*coeff_alpha;
                                
            end
            
             
             if isempty(r.coefficient_robust)                    
                  r.coefficient_robust =   cell_coeff_robust;
                  r.covariance_coefficient_robust = covariance_matrix;
             end
            
        end
        

        function res = erreur(r,graphe,ordre)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%DESCRIPTION: % Return the learning error (An estimation of the mean square error committed by bS)
%USAGE:   res = erreur(bs,graphe,ordre)
%
%INPUT :
%bs               :(bSplines  object) , embed the nonparametric estimators
%OPTIONAL
%graphe           :(integer default 0) integer set to 1 if the estimator and the data has to be plotted 
%OUTPUT:
%res             :(real) estimated mean square error
%               
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

         if nargin <= 2
         ordre = 0;
         end
         
         if nargin <= 1
         graphe = 0;
         end

         Mat_c = r.observation.getMatrix();
         T_c = r.observation.getTimes();
         dim = length(T_c);
         vraie_val = cell(dim,1);
         val_pred = cell(dim,1);
         
         nvT_c = T_c;
   
         if ordre == 1
             for ni = 1 : dim
             T = T_c{ni};
             Mat_res = Mat_c{ni};
             Xn = calcul_prediction(r,T,ordre,ni);
             dt = T(2:length(T)) - T(1:(length(T)-1));
             dMat = Mat_res(2:length(T)) - Mat_res(1:(length(T)-1));
             
             Mat_res = (ones(length(T)-1,1)./dt).*dMat;
             
             Xn = Xn(1:(length(T)-1),:);

             vraie_val{ni} = Mat_res;
             val_pred{ni} = Xn;
             nvT_c{ni} = T(2:length(T));
             end
         else
             for ni = 1 : dim
                 T = T_c{ni};
                 vraie_val{ni} = Mat_c{ni};
                  val_pred{ni} = calcul_prediction(r,T,ordre,ni);
             end
         end
         
         
         for nh = 1: dim
            res_temp = vraie_val{ni} - val_pred{ni};
            res = sum(sum(res_temp.*res_temp))/(length(vraie_val{ni})*length(vraie_val{ni}));
         end
            
            if graphe == 1
                for nj = 1 : dim -1
                 plot(nvT_c{nj},val_pred{nj},'b',nvT_c{nj},vraie_val{nj},'rx');
                 hold on;
                end
                 plot(nvT_c{dim},val_pred{dim},'b',nvT_c{dim},vraie_val{dim},'rx');
                
            end
        end
        
        % Attribut de nouveau noeuds pour la regression et met  jour les
        % valeur de gcv
        %Argument :
        %Bs : BSplines,base de regression dont on veut changer les noeuds
        %val_knots : cell arrays dont chaque case correspond  un ensemble
        %de noeuds rang dans l'ordre croissant. Le premier ensemble de
        %noeuds correspond  la premire dimension des observations , le
        %deuxime ensemble  la deuxime dimension etc..
        function setKnots(Bs,val_knots,val_ini)
           longueur =  length(Bs.valeur_gcv);
         
         if nargin < 2+1
             val_ini = [];
         end
         
         if length(val_knots) == longueur
             
             C = cell(1,longueur);
             D = cell(1,longueur);
             E = cell(1,longueur);
            M_res =  Bs.observation.getMatrix();
            T_c = Bs.observation.getTimes();
            
                 for i = 1 : longueur
                     T = T_c{i};
                     M_value = M_res{i};
                     Bn  = bsplineM(T,val_knots{i},Bs.ordre +1,0,0);
                     M_aux = pinv((Bn')*Bn)*(Bn');
                     H = Bn*M_aux;
                     
          if ~isempty(val_ini)             
            A_eq = zeros(1,length(Bn(1,:)));
            A_eq(1,1) = 1;
            val_constraint_ini = val_ini(i);
            Mat_E = lsqlin(Bn,M_value,[],[],A_eq,val_constraint_ini);
          else
             Mat_E = M_aux*M_value;    
          end                                                                                 
                     aux_gcv = sum((M_value - H*M_value).*(M_value - H*M_value));
                     Gcv = aux_gcv/(length(T)*(1 - trace(H)/length(T))^2);
                     C(i) = val_knots(i);
                     D(i) = {Gcv};
                     E(i) = {Mat_E};
                     Bs.mat_inter{i} = Bn;
                 end
             
            Bs.coefficient = E;
            Bs.knots = C;
            Bs.valeur_gcv = D;

         end
        
        end
    
        %Renvoie un tableau de cell  array de taille egale aux nombre de
        %variable d'tat (plus une matrice)
        function Cn =  getH0(Bs)
            Obs = Bs.getObservation();
            T_c = Obs.getTimes();
            C= Bs.knots;
            d = length(C);
            %Cn  = zeros(d,length(T));
            Cn = cell(d,1);
            p = Bs.ordre;
            for j = 1 : d
              T = T_c{j};  
              Bn  = bsplineM(T,C{1,j},p +1,0,0);
             % size(C{1,j})
              Cn{j}= bsplineM(T(1),C{1,j},p +1,0,0)*pinv(Bn'*Bn)*Bn';
            end
        end
        
        function Ln = getValeurLisseur(Bs,T,nb_obs,ordre_der)
            
            if nargin <= 2+1
                ordre_der = 0;
            end
          Bn =   Bs.mat_inter{nb_obs};
          Mat_aux =  pinv(Bn'*Bn)*Bn';
          Ln =  ((bsplineM(T,Bs.knots{nb_obs},Bs.ordre +1,ordre_der,0))* Mat_aux)';
        end
        
        function Pn = getValeurProjection(r,x,j,order)
            if nargin <= 3
                order = 0;
            end
            
        Pn = bsplineM(x,r.knots{j},r.ordre+1,order,0);       
        end
        
        %Estime la variance du bruit sur les donnes
        %arguments:
        %Bs : estimateur non-paramtrique des courbes
        %j : indice de la variable d'tat correspondant aux donnes
        % type : entier correspondant au type d'estimateur de la variance
            % 1 (defaut) estimateur dans le cadre d'un bruit homoscedastique
            % 2 estimateur dans le cadre d'un bruit heteroscedastique
            % 3 estimateur dans le cadre d'un bruit heteroscedastique
            % 4 estimateur dans le cadre d'autocorrelation (hac)
        
        function val = getValeurVariance(Bs,j,type)
            
            if nargin <= 2
                type = 1;
            end
            
            Obs = Bs.getObservation();
            T_c = Obs.getTimes();
            n_j = length(T_c{j});
            Matrix =  Obs.getMatrix();
            
            Mat_X = Bs.mat_inter{j};
            X_inv = pinv(Mat_X'*Mat_X);
            
            if type == 1
           Rn = calcul_prediction(Bs,T_c{j},0,j);     
           somme_aux = (Matrix{j} - Rn).^2;
           somme = sum(somme_aux);
           Bn =   Bs.mat_inter{j};
           H = Bn*pinv(Bn'*Bn)*Bn';
           v = trace(H);
           v_tilde = trace((H')*H);
           
           val = somme/(n_j -2*v + v_tilde);
           
           val = val.*X_inv;
            end
            
            if type == 2
                Rn = calcul_prediction(Bs,T_c{j},0,j);
                square_residual_u = (Matrix{j}  - Rn).^2;
                val = X_inv*(Mat_X')*diag(square_residual_u)*Mat_X*X_inv;
            end
            
            if type == 3
               Rn = calcul_prediction(Bs,T_c{j},0,j);     
               square_residual_u = (Matrix{j}  - Rn).^2;
               Bn = Bs.mat_inter{j};
               Hat  = Bn*pinv((Bn')*Bn)*(Bn');
               denom = (1 - diag(Hat)).^2;
               
               %indice_denom_non_nul = find((1-denom<=0));
               indice_residu_nul = find(square_residual_u<=0);
               quotient = square_residual_u./denom;
               quotient(indice_residu_nul) = zeros(length(indice_residu_nul));
               
               val = X_inv*(Mat_X')*diag(quotient)*Mat_X*X_inv;
            end
            
            if type == 4
                Rn = calcul_prediction(Bs,T_c{j},0,j);   
                Bn = Bs.mat_inter{j};
                residual = (Matrix{j}  - Rn);
                [ligne col] = size(Bn');
                nb_obs = length(T_c{j});
                func_estim = zeros(ligne,col);
                for gh = 1 : nb_obs
                    func_estim(:,gh) = (Bn(gh,:).*residual(gh))';
                end
                
                val_temp = zeros(ligne);
                for ngi = 1: nb_obs
                    for ngj = 1 : nb_obs
                        coeff = 1 - abs(ngj-ngi)/nb_obs;
                    val_temp = val_temp + coeff.*func_estim(:,ngi)*func_estim(:,ngj)';    
                    end
                end
                
                val = X_inv'*val_temp*X_inv;
                
            end
        end
        
         %Calcul et trace pour les dates d'observations l'intervalle de
        %confiance de l'estimateur non-paramtrique de la variable desir
        %Argument
        % Bs : lisseur linaire utilis
        % var_etat : indice de la variable d'tat dont on veut l'intervalle
        % de confiance
        % alpha : correspond  l'intervalle de confiance (1-alpha)% 
        %est_type : type d'estimateur de la variance de l'estimateur non
        %parametrique des courbes (par dfaut 2 (htroscedastique))
        function [val1 val2] = Confidence_set(Bs,var_etat,alpha,est_type)
            
            if nargin < 3
                alpha = 0.05;
            end
            
            if nargin < 4
                est_type = 2;
            end
            
            variance = getValeurVariance(Bs,var_etat,est_type);

            c = valeur_c(Bs,var_etat,alpha);
            Obs = Bs.getObservation();
            T_c = Obs.getTimes();
            T = T_c{var_etat};
            Ln = getValeurLisseur(Bs,T,var_etat,0);
            Rn = calcul_prediction(Bs,T,0,var_etat);
            

           % variance
         

          val1 = Rn - (c*diag(variance)'.*arrayfun(@sqrt,sum(Ln.^2)))';
          val2 = Rn + (c*diag(variance)'.*arrayfun(@sqrt,sum(Ln.^2)))';
          
          plot(T,Rn,'r',T,val1,'b',T,val2,'g');
          %plot(T,Rn,'r');
            
%             function res = val_inf(x)
%             val_1_aux = getValeurLisseur(Bs,x,var_etat,0);
%            res = calcul_prediction(Bs,x0) - c*variance*sqrt(val_1_aux.^2);
%             end
%             
%             function res = val_sup(x)
%             val_1_aux = getValeurLisseur(Bs,x,var_etat,0);
%            res = calcul_prediction(Bs,x0) - c*variance*sqrt(val_1_aux.^2);
%             end
%             
%             val1 = @val_inf;
%             val2 = @val_sup;
        end
        
          function setValeurk0(Bs,val_ko,index)
             Bs.valeur_k(index) = val_ko;
         end
         
         function res = getValeurk0(Bs,index)
            res = Bs.valeur_k(index) ;
         end
    
    end
end
