function [est_param,fval] = est_param_nls_autonomous_linear(Obs,mat_A_theta,der_mat_AX_theta,dim_syst,param_ini,x_0,type_solver)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%DESCRIPTION: % Plot the economic function corresponding to the pontryagin estimator for non linear or linear model
%
%USAGE:   res_fval = plot_func_eco_pontryagin_general_case(bs,vect_field,der_vect_field,der_sec_vect_field,mat_Obs,mat_U,param_trial,x_0,param_fixe,type_solver)
%
%INPUT :
%bs               :(bSplines  object) , embed the nonparametric estimators
%vect_field       :(function handler) embed the function corresponding to the vector field with signature: 
%                  res_t =  vect_field(t,vx,param,param_fixe)
%                           t : time
%                           vx : state value at time t
%                           param : varying parameter value for the plotted graph
%                           param_fixe : other parameter value
%                           res_t : (vector of size dim x 1) vector field  value at time t
%der_vect_field       :(function handler) embed the function corresponding to the vector field derivative form x.rt state with signature: 
%                  res_t = der_vect_field(t,vx,param,param_fixe)
%                           t : time
%                           vx : state value at time t
%                           param : varying parameter value for the plotted graph
%                           param_fixe : other parameter value
%                           res_t : (matrix of size dim x dim) matrix der_f_i/x_j: vector field derivative form x.rt state value at time t
%der_sec_vect_field       :(function handler) embed the function corresponding to the vector field second derivative form x.rt state with signature: 
%                  res_t = der_sec_vect_field(t,vx,param,param_fixe)
%                           t : time
%                           vx : state value at time t
%                           param : varying parameter value for the plotted graph
%                           param_fixe : other parameter value
%                           res_t : (hypermatrix of size dim x dim x dim) matrix der_sec_f_i/x_j_x_h: vector field second derivative form x.rt state value at time t
%mat_Obs                  : (matrix of size dim_obs x dim): observation matrix C
%mat_U                    : (matrix of size dim x dim): ponderation matrix U for control
%param_trial              : (matrix of size 1 x 2 or 2 x 2) lower and upper bound for varying parameter(s) value
%OPTIONAL
%x_0                      :(vector of size 1 x dim) initial condition if known (default empty ([]) )
%param_fixe               :(vector) fixed parameter for the model (default empty ([]) )
%type_solver              : (integer) indicate the used mixed boundary ode solver
%                         0 (default) extremal variation algorithm
%                         1           quasilinearisation algorithm
%
%OUTPUT:
%res_fval                 :(matrix or vector) economical function value for each varying parameter(s) value
%               
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

    T_c = Obs.getTimes();
     Times = T_c{1};
    Data_c= Obs.getMatrix();
    Data_mat = cell2mat(Data_c');
    Data_mat = reshape(Data_mat',length(Times)*dim_syst,1);
    nb_param = length(param_ini);    

    pas_solve =200;
    Int_ode_solve = Times(1):((Times(end)-Times(1))/pas_solve):Times(end);
    x_0 = reshape(x_0,dim_syst,1);

     
    if nargin < 7
        type_solver = 1;
    end

     
    function [res_in res_eq] = mycon(param)
        res_in = 0;
        res_eq = 0;
    end
 
    
      function [res_val res_grad]= func_eco_multi_param(param)
          %param
      
          function dy = equ_sens_b(t,y)
                val_sens_x = reshape(y,dim_syst,nb_param); 
                res_x_t = expm(mat_A_theta(param)*t)*x_0;
                val_r_theta = der_mat_AX_theta(param,res_x_t);
                dy = mat_A_theta(param)*val_sens_x +  val_r_theta;
                dy = reshape(dy,nb_param*dim_syst,1);
          end
          
                  
          function res_x_t = val_x(t)
              res_x_t = expm(mat_A_theta(param)*t)*x_0;
          end
 
          sol_equ_sens = ode45(@equ_sens_b,Int_ode_solve,zeros(nb_param*dim_syst,1)); 

          function res_grad_t = val_sens_x(t)
             res_grad_t = reshape(deval(sol_equ_sens,t),dim_syst,nb_param);
          end
          
          res_x_T =  arrayfun(@val_x,Times,'UniformOutput',false);
          res_x_T = cell2mat(res_x_T);
          res_sens_x_T =  arrayfun(@val_sens_x,Times,'UniformOutput',false);
          res_sens_x_T = cell2mat(res_sens_x_T);
          
          residual = (res_x_T - Data_mat);
          res_val = (residual')*residual;
          res_grad = 2*(res_sens_x_T')*residual;
      end
  

           if type_solver == 1
              'fminunc'  
            nb_fun_eval = 1000;
            max_iter = 2000;
            options = optimset('TolFun',1e-8,'TolX',1e-8,'GradObj','on','MaxFunEvals',nb_fun_eval,'MaxIter',max_iter,'Hessian','off');
            [est_param,fval]= fminunc(@func_eco_multi_param,param_ini,options);
           else
                             
                    max_iter = 100;
                   nb_fun_eval = 1000; 
                   options = optimset('TolFun',1e-7,'TolX',1e-7,'GradObj','on','MaxFunEvals',nb_fun_eval,'MaxIter',max_iter);
                   
                   lb = zeros(length(param_ini),1);
                   ub = zeros(length(param_ini),1);
                   ub(1:length(param_ini)) = +Inf;
                   
                   if type_solver == 3
                        'fmincon Mat_A'
                       Mat_A = -eye(length(param_ini));
                       vect_b = zeros(length(param_ini),1);
                       [est_param,fval] =  fmincon(@func_eco_multi_param,param_ini,Mat_A,vect_b,[],[],lb,ub,@mycon,options)
                   else
                        'fmincon bound'
                        [est_param,fval] =  fmincon(@func_eco_multi_param,param_ini,[],[],[],[],lb,ub,@mycon,options)
                   end

       
%         nb_fun_eval = 100000;
%         max_iter = 1000;
%         options = optimset('TolFun',1e-10,'TolX',1e-10,'GradObj','off','MaxFunEvals',nb_fun_eval,'MaxIter',max_iter,'Hessian','off');
%         [est_param,fval]= fminunc(@func_eco_multi_param,param_ini,options);
      end


end